@@ -12,8 +12,16 @@ from ._impl.model import (
     Unit,
     FileUnit,
     RpmUnit,
+    RpmDependency,
+    ErratumUnit,
+    ErratumReference,
+    ErratumPackageCollection,
+    ErratumPackage,
+    ErratumModule,
     ModulemdUnit,
+    ModulemdDependency,
     ModulemdDefaultsUnit,
+    YumRepoMetadataFileUnit,
     Distributor,
     PublishOptions,
     FileSyncOptions,
