@@ -1,24 +1,33 @@
-import os
-import logging
-import threading
 import hashlib
 import json
+import logging
+import os
+import threading
 from functools import partial
-import six
-from six.moves import StringIO
 
 import requests
+import six
 from more_executors import Executors
 from more_executors.futures import f_map, f_flat_map, f_return, f_proxy, f_sequence
+from six.moves import StringIO
 
 from ..page import Page
 from ..criteria import Criteria
-from ..model import Repository, MaintenanceReport, Distributor, Unit
+from ..model import (
+    Repository,
+    FileRepository,
+    MaintenanceReport,
+    Distributor,
+    Unit,
+    Task,
+)
 from .search import search_for_criteria
 from .errors import PulpException
 from .poller import TaskPoller
 from . import retry
 
+from .ud_mappings import compile_ud_mappings
+
 
 LOG = logging.getLogger("pubtools.pulplib")
 
@@ -157,6 +166,9 @@ class Client(object):
 
         self._tls = threading.local()
 
+        # Cache supported types, used during searches.
+        self._server_type_ids = None
+
         # executor only for issuing HTTP requests to Pulp:
         # - does not cover watching Pulp tasks
         # - checks HTTP response and raises if it's not JSON or
@@ -253,8 +265,95 @@ class Client(object):
 
         .. versionadded:: 2.6.0
         """
+        # Criteria will be serialized into a Pulp search at the time we
+        # actually do the query, but validate eagerly as well so we raise
+        # ASAP on invalid input.
+        search_for_criteria(criteria, Unit, None)
 
-        server_type_ids = self.get_content_type_ids()
+        if self._server_type_ids is None:
+            # We'll be using this in a moment either to set default IDs or to
+            # reject searches for invalid types.
+            # Note: no locking, so if we're called from multiple threads we in
+            # theory might waste some time querying the types more than once.
+            self._server_type_ids = self.get_content_type_ids()
+
+        return f_proxy(
+            f_flat_map(
+                self._server_type_ids,
+                lambda ids: self._search_content_with_server_type_ids(criteria, ids),
+            )
+        )
+
+    def copy_content(self, from_repository, to_repository, criteria=None):
+        """Copy content from one repository to another.
+
+        Args:
+            from_repository (:class:`~pubtools.pulplib.Repository`)
+                A repository used as the source of content.
+
+            to_repository (:class:`~pubtools.pulplib.Repository`)
+                A repository to receive copied content.
+
+            criteria (:class:`~pubtools.pulplib.Criteria`)
+                A criteria object used to find units for copy.
+                If None, all units in the source repo will be copied.
+
+        Returns:
+            Future[list[:class:`~pubtools.pulplib.Task`]]
+                A future which is resolved when the copy completes.
+
+                The :meth:`~pubtools.pulplib.Task.units` attribute may be inspected
+                to determine which unit(s) were copied. Note that the returned units
+                typically will have only a subset of available fields.
+
+        .. versionadded:: 2.17.0
+        """
+        return f_proxy(
+            self._do_associate(from_repository.id, to_repository.id, criteria)
+        )
+
+    def update_content(self, unit):
+        """Update mutable fields on an existing unit.
+
+        Args:
+            unit (:class:`~pubtools.pulplib.Unit`)
+                A unit to be updated.
+
+                This unit must have a known ``unit_id``.
+
+                Only those fields documented as *mutable* will have any effect during
+                the update (e.g. ``cdn_path``). Other fields either cannot be updated
+                at all, or can only be updated by re-uploading the associated content.
+
+        Returns:
+            Future
+                A future which is resolved with a value of ``None`` once the unit
+                has been updated.
+
+        .. versionadded:: 2.20.0
+        """
+
+        if not unit.unit_id:
+            raise ValueError("unit_id missing on call to update_content()")
+
+        url = os.path.join(
+            self._url,
+            "pulp/api/v2/content/units/%s/%s/pulp_user_metadata/"
+            % (unit.content_type_id, unit.unit_id),
+        )
+
+        out = self._request_executor.submit(
+            self._do_request, method="PUT", url=url, json=unit._usermeta
+        )
+
+        # The Pulp API is defined as returning 'null' so this should be a no-op,
+        # but to be extra sure we don't return anything unexpected, we'll force
+        # the return value to None.
+        out = f_map(out, lambda _: None)
+
+        return out
+
+    def _search_content_with_server_type_ids(self, criteria, server_type_ids):
         prepared_search = search_for_criteria(criteria, Unit, None)
         type_ids = prepared_search.type_ids
         if not type_ids:
@@ -266,7 +365,10 @@ class Client(object):
             )
 
         return self._search(
-            Unit, ["content/units/%s" % x for x in type_ids], criteria=criteria
+            Unit,
+            ["content/units/%s" % x for x in type_ids],
+            criteria=criteria,
+            search_options={"include_repos": True},
         )
 
     def search_distributor(self, criteria=None):
@@ -288,6 +390,25 @@ class Client(object):
         """
         return self._search(Distributor, "distributors", criteria=criteria)
 
+    def search_task(self, criteria=None):
+        """Search the tasks matching the given criteria.
+
+        Args:
+            criteria (:class:`~pubtools.pulplib.Criteria`)
+                A criteria object used for this search.
+                If None, search for all tasks.
+
+        Returns:
+            Future[:class:`~pubtools.pulplib.Page`]
+                A future representing the first page of results.
+
+                Each page will contain a collection of
+                :class:`~pubtools.pulplib.Task` objects.
+
+        .. versionadded:: 2.19.0
+        """
+        return self._search(Task, "tasks", criteria=criteria)
+
     def _search(
         self,
         return_type,
@@ -415,7 +536,7 @@ class Client(object):
         # We only support returning the ID at this time.
         return f_proxy(f_map(out, lambda types: sorted([t["id"] for t in types])))
 
-    def _do_upload_file(self, upload_id, file_obj, name):
+    def _do_upload_file(self, upload_id, file_obj):
         def do_next_upload(checksum, size):
             data = file_obj.read(self._CHUNK_SIZE)
             if data:
@@ -434,15 +555,16 @@ class Client(object):
         if not is_file_object:
             file_obj = open(file_obj, "rb")
 
-        LOG.info("Uploading %s to Pulp", name)
         upload_f = f_flat_map(f_return(), lambda _: do_next_upload(hashlib.sha256(), 0))
 
-        if not is_file_object:
-            upload_f.add_done_callback(lambda _: file_obj.close())
+        upload_f.add_done_callback(lambda _: file_obj.close())
+
         return upload_f
 
     def _publish_repository(self, repo, distributors_with_config):
-        tasks_f = f_return([])
+        compiled = self._compile_notes(repo)
+
+        tasks_f = f_map(compiled, lambda _: [])
 
         def do_next_publish(accumulated_tasks, distributor, config):
             distributor_tasks_f = self._publish_distributor(
@@ -470,14 +592,36 @@ class Client(object):
             self._do_request, method="POST", url=url, json=body
         )
 
-    def _do_unassociate(self, repo_id, type_ids):
+    def _do_associate(self, src_repo_id, dest_repo_id, criteria=None):
+        url = os.path.join(
+            self._url, "pulp/api/v2/repositories/%s/actions/associate/" % dest_repo_id
+        )
+
+        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)
+
+        body = {"source_repo_id": src_repo_id, "criteria": {}}
+        if pulp_search.type_ids:
+            body["criteria"]["type_ids"] = pulp_search.type_ids
+        if pulp_search.filters:
+            body["criteria"]["filters"] = {"unit": pulp_search.filters}
+
+        LOG.debug("Submitting %s associate: %s", url, body)
+
+        return self._task_executor.submit(
+            self._do_request, method="POST", url=url, json=body
+        )
+
+    def _do_unassociate(self, repo_id, criteria=None):
         url = os.path.join(
             self._url, "pulp/api/v2/repositories/%s/actions/unassociate/" % repo_id
         )
 
-        body = {}
-        if type_ids is not None:
-            body["type_ids"] = type_ids
+        # use type hint=Unit so that if type_ids are the goal here
+        # then we will get back a properly prepared PulpSearch with
+        # a populated type_ids field
+        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)
+
+        body = {"criteria": {"type_ids": pulp_search.type_ids}}
 
         LOG.debug("Submitting %s unassociate: %s", url, body)
 
@@ -485,6 +629,25 @@ class Client(object):
             self._do_request, method="POST", url=url, json=body
         )
 
+    def _compile_notes(self, repo):
+        # Given a repo we're about to publish, calculate and set certain notes
+        # derived from the repo contents.
+        out = f_return()
+
+        if isinstance(repo, FileRepository):
+            out = compile_ud_mappings(
+                repo,
+                # UD mapping code is written to do low-level pulp requests.
+                # In order to keep the code at arms length and don't let it access
+                # client internals directly, we'll just pass a callback
+                # it can use to do these requests.
+                do_request=lambda url, **kwargs: self._request_executor.submit(
+                    self._do_request, url=os.path.join(self._url, url), **kwargs
+                ),
+            )
+
+        return out
+
     @classmethod
     def _unpack_response(cls, pulp_response):
         try:
@@ -614,15 +777,20 @@ class Client(object):
             self._do_request, method="PUT", url=url, data=data
         )
 
-    def _do_import(self, repo_id, upload_id, unit_type_id, unit_key):
+    def _do_import(
+        self, repo_id, upload_id, unit_type_id, unit_key, unit_metadata=None
+    ):
         url = os.path.join(
             self._url, "pulp/api/v2/repositories/%s/actions/import_upload/" % repo_id
         )
 
+        unit_metadata = unit_metadata or {}
+
         body = {
             "unit_type_id": unit_type_id,
             "upload_id": upload_id,
             "unit_key": unit_key,
+            "unit_metadata": unit_metadata,
         }
 
         LOG.debug("Importing contents to repo %s with upload id %s", repo_id, upload_id)
