@@ -2,13 +2,14 @@ import datetime
 import logging
 
 from attr import validators, asdict
-from more_executors.futures import f_proxy, f_map
+from frozenlist2 import frozenlist
+from more_executors.futures import f_proxy, f_map, f_flat_map
 
-from ..common import PulpObject, Deletable, DetachedException
 from ..attr import pulp_attrib
+from ..common import PulpObject, Deletable, DetachedException
+from ..convert import frozenlist_or_none_converter
 from ..distributor import Distributor
-from ..frozenlist import FrozenList
-from ...criteria import Criteria
+from ...criteria import Criteria, Matcher
 from ...schema import load_schema
 from ... import compat_attr as attr
 from ...hooks import pm
@@ -58,6 +59,15 @@ class PublishOptions(object):
     :meth:`~pubtools.pulplib.Distributor.is_rsync` is ``True``.
     """
 
+    rsync_extra_args = pulp_attrib(
+        default=None, type=list, converter=frozenlist_or_none_converter
+    )
+    """If present, provide these additional arguments to any rsync commands run
+    during publish.
+
+    Ignored when rsync is not used.
+    """
+
 
 @attr.s(kw_only=True, frozen=True)
 class SyncOptions(object):
@@ -112,6 +122,14 @@ class SyncOptions(object):
     """A string representing the password that should be used to authenticate with the proxy server
     """
 
+    basic_auth_username = pulp_attrib(default=None, type=str)
+    """Username to authenticate with source which supports basic authentication.
+    """
+
+    basic_auth_password = pulp_attrib(default=None, type=str)
+    """Password to authenticate with source which supports basic authentication.
+    """
+
 
 @attr.s(kw_only=True, frozen=True)
 class Repository(PulpObject, Deletable):
@@ -147,11 +165,11 @@ class Repository(PulpObject, Deletable):
     """
 
     distributors = pulp_attrib(
-        default=attr.Factory(FrozenList),
+        default=attr.Factory(frozenlist),
         type=list,
         pulp_field="distributors",
-        converter=FrozenList,
-        pulp_py_converter=lambda ds: FrozenList([Distributor.from_data(d) for d in ds]),
+        converter=frozenlist,
+        pulp_py_converter=lambda ds: frozenlist([Distributor.from_data(d) for d in ds]),
         # It's too noisy to let repr descend into sub-objects
         repr=False,
     )
@@ -172,7 +190,7 @@ class Repository(PulpObject, Deletable):
     """Default publish URL for this repository, relative to the Pulp content root."""
 
     mutable_urls = pulp_attrib(
-        default=attr.Factory(FrozenList), type=list, converter=FrozenList
+        default=attr.Factory(frozenlist), type=list, converter=frozenlist
     )
     """A list of URLs relative to repository publish root which are expected
     to change at every publish (if any content of repo changed)."""
@@ -198,12 +216,12 @@ class Repository(PulpObject, Deletable):
     """
 
     signing_keys = pulp_attrib(
-        default=attr.Factory(FrozenList),
+        default=attr.Factory(frozenlist),
         type=list,
         pulp_field="notes.signatures",
         pulp_py_converter=lambda sigs: sigs.split(","),
         py_pulp_converter=",".join,
-        converter=lambda keys: FrozenList([k.strip() for k in keys]),
+        converter=lambda keys: frozenlist([k.strip() for k in keys]),
     )
     """A list of GPG signing key IDs used to sign content in this repository."""
 
@@ -374,7 +392,16 @@ class Repository(PulpObject, Deletable):
         if not self._client:
             raise DetachedException()
 
-        # all distributor IDs we're willing to invoke. Anything else is ignored.
+        # Before adding distributors and publishing, we'll activate this hook
+        # to allow subscribing implementers the opportunity to adjust options.
+        hook_rets = pm.hook.pulp_repository_pre_publish(
+            repository=self, options=options
+        )
+        # Use the first non-None hook return value to replace options.
+        hook_rets = [ret for ret in hook_rets if ret is not None]
+        options = hook_rets[0] if hook_rets else options
+
+        # All distributor IDs we're willing to invoke. Anything else is ignored.
         # They'll be invoked in the order listed here.
         candidate_distributor_ids = self._PUBLISH_DISTRIBUTORS
 
@@ -471,8 +498,31 @@ class Repository(PulpObject, Deletable):
         # version of this method will support some "criteria".  Let's not fix the
         # argument order at least until then.
 
+        # start down the path of using Criteria per this issue:
+        # https://github.com/release-engineering/pubtools-pulplib/issues/62
+        # by using Criteria internally.
+        # there seems to be pretty complex handling of Criteria
+        # for serialization in the search API, and it is unclear which parts
+        # might also be necessary to use for content removal.
+        # If any or all of the same handling is needed, it would be beneficial
+        # to encapsulate the preparation of a criteria JSON object in some
+        # (more generically named) functions or a class to avoid duplicating code.
+        # for reference see search_content, _impl.client.Client._search_repo_units,
+        # _impl.client.Client._search, and _impl.client.search.search_for_criteria
+        criteria = None
         type_ids = kwargs.get("type_ids")
-        return f_proxy(self._client._do_unassociate(self.id, type_ids))
+        # use _content_type_id field name to coerce
+        # search_for_criteria to fill out the PulpSearch#type_ids field.
+        # passing a criteria with an empty type_ids list rather than
+        # None results in failing tests due to the implementation of
+        # FakeClient#_do_unassociate
+        if type_ids is not None:
+            criteria = Criteria.with_field(
+                "_content_type_id",
+                Matcher.in_(type_ids),  # Criteria.with_field_in is deprecated
+            )
+
+        return f_proxy(self._client._do_unassociate(self.id, criteria=criteria))
 
     @classmethod
     def from_data(cls, data):
@@ -504,11 +554,13 @@ class Repository(PulpObject, Deletable):
     def _config_for_distributor(cls, distributor, options):
         out = {}
 
-        if options.clean is not None and distributor.is_rsync:
-            out["delete"] = options.clean
-
-        if options.origin_only is not None and distributor.is_rsync:
-            out["content_units_only"] = options.origin_only
+        if distributor.is_rsync:
+            if options.clean is not None:
+                out["delete"] = options.clean
+            if options.origin_only is not None:
+                out["content_units_only"] = options.origin_only
+            if options.rsync_extra_args is not None:
+                out["rsync_extra_args"] = options.rsync_extra_args
 
         if options.force is not None:
             out["force_full"] = options.force
@@ -521,3 +573,86 @@ class Repository(PulpObject, Deletable):
         # distributors use the same client as owning repository
         for distributor in self.distributors or []:
             distributor._set_client(client)
+
+    def _upload_then_import(
+        self, file_obj, name, type_id, unit_key_fn=None, unit_metadata_fn=None
+    ):
+        """Private helper to upload and import a piece of content into this repo.
+
+        To be called by the type-specific subclasses (e.g. YumRepository,
+        FileRepository...)
+
+        Args:
+            file_obj (str, file-like object, None):
+                file object or path (as documented in public methods), or None
+                if this unit type has no associated file
+
+            name (str):
+                a brief user-meaningful name for the content being uploaded
+                (appears in logs)
+
+            type_id (str):
+                pulp unit type ID
+
+            unit_key_fn (callable):
+                a callable which will be invoked with the return value of
+                _do_upload_file (or None if file_obj is None).
+                It should return the unit key for this piece of
+                content. If omitted, an empty unit key is used, which means Pulp
+                is wholly responsible for calculating the unit key.
+
+            unit_metadata_fn (callable):
+                a callable which will be invoked with the return value of
+                _do_upload_file (or None if file_obj is None). It should return
+                the unit metadata for this piece of
+                content. If omitted, metadata is not included in the import call to
+                Pulp.
+        """
+
+        if not self._client:
+            raise DetachedException()
+
+        unit_key_fn = unit_key_fn or (lambda _: {})
+        unit_metadata_fn = unit_metadata_fn or (lambda _: None)
+
+        upload_id_f = f_map(
+            self._client._request_upload(), lambda upload: upload["upload_id"]
+        )
+
+        f_map(
+            upload_id_f,
+            lambda upload_id: LOG.info(
+                "Uploading %s to %s [%s]", name, self.id, upload_id
+            ),
+        )
+
+        if file_obj is None:
+            # If there is no file for this kind of unit (e.g. erratum),
+            # we still have to use the request_upload and import APIs; we just
+            # never upload any bytes. That means the upload is 'complete' as
+            # soon as the upload ID is known. A real upload returns a (size, checksum)
+            # tuple; we force a no-content upload to return None.
+            upload_complete_f = f_map(upload_id_f, lambda _: None)
+        else:
+            upload_complete_f = f_flat_map(
+                upload_id_f,
+                lambda upload_id: self._client._do_upload_file(upload_id, file_obj),
+            )
+
+        import_complete_f = f_flat_map(
+            upload_complete_f,
+            lambda upload: self._client._do_import(
+                self.id,
+                upload_id_f.result(),
+                type_id,
+                unit_key_fn(upload),
+                unit_metadata_fn(upload),
+            ),
+        )
+
+        f_map(
+            import_complete_f,
+            lambda _: self._client._delete_upload_request(upload_id_f.result()),
+        )
+
+        return f_proxy(import_complete_f)
